# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 European Union.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from flask_principal import AnonymousIdentity
from invenio_access.permissions import any_user as any_user_need
from invenio_accounts.proxies import current_datastore
from invenio_app.factory import create_api

from invenio_users_resources.proxies import (
    current_groups_service,
    current_users_service,
)
from invenio_users_resources.records import GroupAggregate, UserAggregate

pytest_plugins = ("celery.contrib.pytest",)


#
# Application
#
@pytest.fixture(scope="module")
def app_config(app_config):
    """Override pytest-invenio app_config fixture."""
    app_config[
        "RECORDS_REFRESOLVER_CLS"
    ] = "invenio_records.resolver.InvenioRefResolver"
    app_config[
        "RECORDS_REFRESOLVER_STORE"
    ] = "invenio_jsonschemas.proxies.current_refresolver_store"
    # Variable not used. We set it to silent warnings
    app_config["JSONSCHEMAS_HOST"] = "not-used"

    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path):
    """Application factory fixture."""
    return create_api


@pytest.fixture()
def headers():
    """Default headers for making requests."""
    return {
        "content-type": "application/json",
        "accept": "application/json",
    }


#
# Services
#
@pytest.fixture(scope="module")
def user_service(app):
    """User service."""
    return current_users_service


@pytest.fixture(scope="module")
def group_service(app):
    """Group service."""
    return current_groups_service


#
# Users
#
@pytest.fixture(scope="module")
def anon_identity():
    """A new user."""
    identity = AnonymousIdentity()
    identity.provides.add(any_user_need)
    return identity


@pytest.fixture(scope="module")
def users_data():
    """Data for users."""
    return [
        {
            "username": "pubres",
            "email": "pubres@inveniosoftware.org",
            "profile": {
                "full_name": "Tim Smith",
                "affiliations": "CERN",
            },
            "preferences": {
                "visibility": "public",
                "email_visibility": "restricted",
            },
        },
        {
            "username": "pub",
            "email": "pub@inveniosoftware.org",
            "profile": {
                "full_name": "Jose Benito Gonzalez Lopez",
                "affiliations": "CERN",
            },
            "preferences": {
                "visibility": "public",
                "email_visibility": "public",
            },
        },
        {
            "username": "res",
            "email": "res@inveniosoftware.org",
            "profile": {
                "full_name": "Donat Agosti",
                "affiliations": "Plazi",
            },
            "preferences": {
                "visibility": "restricted",
                "email_visibility": "restricted",
            },
        },
        {
            "username": "unconfirmed",
            "email": "unconfirmed@inveniosoftware.org",
            "confirmed": False,
        },
        {
            "username": "inactive",
            "email": "inactive@inveniosoftware.org",
            "profile": {
                "full_name": "Spammer",
                "affiliations": "Spam org",
            },
            "preferences": {
                "visibility": "public",
                "email_visibility": "public",
            },
            "active": False,
        },
    ]


@pytest.fixture(scope="module")
def users(UserFixture, app, database, users_data):
    """Test users."""
    users = {}
    for obj in users_data:
        u = UserFixture(
            username=obj["username"],
            email=obj["email"],
            password=obj["username"],
            user_profile=obj.get("profile"),
            preferences=obj.get("preferences"),
            active=obj.get("active", True),
            confirmed=obj.get("confirmed", True),
        )
        u.create(app, database)
        users[obj["username"]] = u
    UserAggregate.index.refresh()
    return users


def _create_group(name, description, database):
    """Creates a Role/Group."""
    r = current_datastore.create_role(name=name, description=description)
    current_datastore.commit()

    return r


@pytest.fixture(scope="module")
def group(database):
    """A single group."""
    r = _create_group(name="it-dep", description="IT Department", database=database)

    GroupAggregate.index.refresh()
    return r


@pytest.fixture(scope="module")
def groups(database, group):
    """A single group."""
    roles = [group]  # it-dep
    roles.append(
        _create_group(name="hr-dep", description="HR Department", database=database)
    )

    GroupAggregate.index.refresh()
    return roles


@pytest.fixture(scope="module")
def user_pub(users):
    """User jbenito (restricted/restricted)."""
    return users["pub"]


@pytest.fixture(scope="module")
def user_pubres(users):
    """User tjs (public/restricted)."""
    return users["pubres"]


@pytest.fixture(scope="module")
def user_res(users):
    """User agosti (restricted/restricted)."""
    return users["res"]


@pytest.fixture(scope="module")
def user_inactive(users):
    """Inactive user."""
    return users["inactive"]


@pytest.fixture(scope="module")
def user_unconfirmed(users):
    """Unconfirmed user."""
    return users["unconfirmed"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 KTH Royal Institute of Technology
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 European Union.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Groups service."""

from invenio_accounts.models import Role
from invenio_records_resources.resources.errors import PermissionDeniedError
from invenio_records_resources.services import RecordService

from ...records.api import GroupAggregate
from ..results import AvatarResult


class GroupsService(RecordService):
    """User groups service."""

    def read(self, identity, id_):
        """Retrieve a user group."""
        # resolve and require permission
        group = GroupAggregate.get_record_by_name(id_)
        if group is None:
            # return 403 even on empty resource due to security implications
            raise PermissionDeniedError()

        self.require_permission(identity, "read", record=group)

        # run components
        for component in self.components:
            if hasattr(component, "read"):
                component.read(identity, group=group)

        return self.result_item(self, identity, group, links_tpl=self.links_item_tpl)

    def read_avatar(self, identity, name_):
        """Get a groups's avatar."""
        group = GroupAggregate.get_record_by_name(name_)
        if group is None:
            # return 403 even on empty resource due to security implications
            raise PermissionDeniedError()
        self.require_permission(identity, "read", record=group)
        return AvatarResult(group)

    def rebuild_index(self, identity, uow=None):
        """Reindex all user groups managed by this service."""
        for role in Role.query.all():
            role_agg = self.record_cls.from_role(role)
            self.indexer.index(role_agg)

        return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Permission generators for users and groups."""

import operator
from functools import reduce
from itertools import chain

from invenio_records_permissions.generators import Generator, UserNeed
from invenio_search.engine import dsl


class IfPublic(Generator):
    """Generator for different permissions based on the visibility settings."""

    def __init__(self, field_name, then_, else_):
        """Constructor."""
        self._field_name = field_name
        self.then_ = then_
        self.else_ = else_

    def _generators(self, record):
        """Get the "then" or "else" generators."""
        if record is None:
            return self.else_

        visibility = record.preferences.get(self._field_name, "restricted")
        is_public = visibility == "public"

        return self.then_ if is_public else self.else_

    def needs(self, record=None, **kwargs):
        """Set of Needs granting permission."""
        needs_chain = chain.from_iterable(
            [g.needs(record=record, **kwargs) for g in self._generators(record)]
        )
        return list(set(needs_chain))

    def excludes(self, record=None, **kwargs):
        """Set of Needs denying permission."""
        needs_chain = chain.from_iterable(
            [g.excludes(record=record, **kwargs) for g in self._generators(record)]
        )
        return list(set(needs_chain))

    def make_query(self, generators, **kwargs):
        """Make a query for one set of generators."""
        queries = [g.query_filter(**kwargs) for g in generators]
        queries = [q for q in queries if q]
        return reduce(operator.or_, queries) if queries else None

    def query_filter(self, **kwargs):
        """Filters for queries."""
        field = f"preferences.{self._field_name}"
        q_public = dsl.Q("match", **{field: "public"})
        q_restricted = dsl.Q("match", **{field: "restricted"})
        then_query = self.make_query(self.then_, **kwargs)
        else_query = self.make_query(self.else_, **kwargs)

        if then_query and else_query:
            return (q_public & then_query) | (q_restricted & else_query)
        elif then_query:
            return q_public & then_query
        elif else_query:
            return q_public | (q_restricted & else_query)
        else:
            return q_public


class IfPublicEmail(IfPublic):
    """Generator checking the 'email_visibility' setting."""

    def __init__(self, then_, else_):
        """Constructor."""
        super().__init__("email_visibility", then_, else_)


class IfPublicUser(IfPublic):
    """Generator checking the 'visibility' setting."""

    def __init__(self, then_, else_):
        """Constructor."""
        super().__init__("visibility", then_, else_)


class Self(Generator):
    """Requires the users themselves."""

    def needs(self, record=None, **kwargs):
        """Set of Needs granting permission."""
        if record is not None:
            return [UserNeed(record.id)]

        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for the current identity."""
        if identity is not None:
            for need in identity.provides:
                if need.method == "id":
                    return dsl.Q("term", id=need.value)

        return []



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User service tests."""

import pytest
from invenio_access.permissions import system_identity


def test_groups_sort(app, groups, group_service):
    # default sort by name
    res = group_service.search(system_identity).to_dict()
    assert res["sortBy"] == "name"
    assert res["hits"]["total"] > 0
    hits = res["hits"]["hits"]
    assert hits[0]["id"] == "hr-dep"
    assert hits[1]["id"] == "it-dep"


def test_groups_no_facets(app, group, group_service):
    """Make sure certain fields ARE searchable."""
    res = group_service.search(system_identity)
    # if facets were enabled but not configured the value would be {}
    assert res.aggregations is None


def test_groups_fixed_pagination(app, groups, group_service):
    res = group_service.search(system_identity, params={"size": 1, "page": 2})
    assert res.pagination.page == 1
    assert res.pagination.size == 10


@pytest.mark.parametrize(
    "query",
    [
        # cannot search on title because is never set
        # see TODO in parse_role_data
        "id:it-dep",
        "name:IT Department",
        "+name:it",
        "IT",
    ],
)
def test_groups_search_field(app, group, group_service, query):
    """Make sure certain fields ARE searchable."""
    res = group_service.search(system_identity, q=query)
    assert res.total > 0



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users service tasks."""

from celery import shared_task
from flask import current_app
from invenio_search.engine import search

from ...proxies import current_groups_service
from ...records.api import GroupAggregate


@shared_task(ignore_result=True)
def reindex_group(role_id):
    """Reindex the given user."""
    index = current_groups_service.record_cls.index
    if current_groups_service.indexer.exists(index):
        try:
            group_agg = GroupAggregate.get_record(role_id)
            current_groups_service.indexer.index(group_agg)
        except search.exceptions.ConflictError as e:
            current_app.logger.warn(f"Could not reindex group {role_id}: {e}")


@shared_task(ignore_result=True)
def unindex_group(role_id):
    """Unindex the given role/group."""
    index = current_groups_service.record_cls.index
    if current_groups_service.indexer.exists(index):
        try:
            group_agg = GroupAggregate.get_record(role_id)
            current_groups_service.indexer.delete(group_agg)
        except search.exceptions.ConflictError as e:
            current_app.logger.warn(f"Could not unindex group {role_id}: {e}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User and user group schemas."""

from flask_babelex import lazy_gettext as _
from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import Schema, ValidationError, fields
from marshmallow_utils.permissions import FieldPermissionsMixin


def validate_visibility(value):
    """Check if the value is a valid visibility setting."""
    if value not in ["public", "restricted"]:
        raise ValidationError(
            message=str(_("Value must be either 'public' or 'restricted'."))
        )


class UserPreferencesSchema(Schema):
    """Schema for user preferences."""

    visibility = fields.String(validate=validate_visibility)
    email_visibility = fields.String(validate=validate_visibility)


class UserProfileSchema(Schema):
    """Schema for user profiles."""

    full_name = fields.String()
    affiliations = fields.String()


class UserSchema(BaseRecordSchema, FieldPermissionsMixin):
    """Schema for users."""

    field_dump_permissions = {
        "email": "read_email",
        "created": "read_details",
        "updated": "read_details",
        "revision_id": "read_details",
        "active": "read_details",
        "confirmed": "read_details",
        "preferences": "read_details",
    }

    # NOTE: API should only deliver users that are active & confirmed
    active = fields.Boolean()
    confirmed = fields.Boolean(dump_only=True)
    is_current_user = fields.Method("is_self", dump_only=True)

    email = fields.String()
    username = fields.String()
    profile = fields.Dict()
    preferences = fields.Nested(UserPreferencesSchema)

    def is_self(self, obj):
        """Determine if identity is the current identity."""
        current_identity = self.context["identity"]
        return (
            obj.id is not None
            and current_identity.id is not None
            and str(obj.id) == str(current_identity.id)
        )


class GroupSchema(BaseRecordSchema):
    """Schema for user groups."""

    name = fields.String()
    title = fields.String()
    description = fields.String()
    provider = fields.String(dump_only=True)
    is_managed = fields.Boolean(dump_only=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""API classes for user and group management in Invenio."""
import random
import unicodedata

from flask import current_app
from invenio_accounts.proxies import current_datastore
from invenio_db import db
from invenio_records.dumpers import SearchDumper
from invenio_records.systemfields import DictField, ModelField
from invenio_records_resources.records.api import Record
from invenio_records_resources.records.systemfields import IndexField

from .dumpers import EmailFieldDumperExt
from .models import GroupAggregateModel, UserAggregateModel


def parse_user_data(user):
    """Parse the user's information into a dictionary."""
    data = {
        "id": user.id,
        "email": user.email,
        "username": user.username,
        "active": user.active,
        "confirmed": user.confirmed_at is not None,
        "preferences": dict(user.preferences or {}),
        "profile": dict(user.user_profile or {}),
    }

    data["preferences"].setdefault("visibility", "restricted")
    data["preferences"].setdefault("email_visibility", "restricted")
    return data


def parse_role_data(role):
    """Parse the role's information into a dictionary."""
    data = {
        "id": role.name,  # due to flask security exposing user id
        "name": role.description,
        "is_managed": True,  # TODO
    }
    return data


class UserAggregate(Record):
    """An aggregate of information about a user."""

    model_cls = UserAggregateModel
    """The model class for the request."""

    # NOTE: the "uuid" isn't a UUID but contains the same value as the "id"
    #       field, which is currently an integer for User objects!
    dumper = SearchDumper(
        extensions=[EmailFieldDumperExt("email")],
        model_fields={"id": ("uuid", int)},
    )
    """Search dumper with configured extensions."""

    metadata = None
    """Disabled metadata field from the base class."""

    index = IndexField("users-user-v1.0.0", search_alias="users")
    """The search engine index to use."""

    # TODO
    id = ModelField("id")
    """The data-layer id."""

    email = DictField("email")
    """The user's email address."""

    username = DictField("username")
    """The user's email address."""

    profile = DictField("profile")
    """The user's profile."""

    active = DictField("active")

    confirmed = DictField("confirmed")

    preferences = DictField("preferences")

    @property
    def avatar_chars(self):
        """Get avatar characters for user."""
        text = None
        if self.profile.get("full_name"):
            text = self.profile["full_name"]
        elif self.username:
            text = self.username
        else:
            alphabet = "ABCDEFGHIJKLMNOPQRSTUVXYZ"
            text = alphabet[self.id % len(alphabet)]

        return text[0].upper()

    @property
    def avatar_color(self):
        """Get avatar color for user."""
        colors = current_app.config["USERS_RESOURCES_AVATAR_COLORS"]
        return colors[self.id % len(colors)]

    @classmethod
    def create(cls, data, id_=None, validator=None, format_checker=None, **kwargs):
        """Create a new User and store it in the database."""
        # NOTE: we don't use an actual database table, and as such can't
        #       use db.session.add(record.model)
        with db.session.begin_nested():
            # create_user() will already take care of creating the profile
            # for us, if it's specified in the data
            user = current_datastore.create_user(**data)
            user_aggregate = cls.from_user(user)
            return user_aggregate

    def _validate(self, *args, **kwargs):
        """Skip the validation."""
        pass

    def commit(self):
        """Update the aggregate data on commit."""
        # TODO this does not allow us to set properties via the UserAggregate?
        #      because everything's taken from the User object...
        data = parse_user_data(self.model.model_obj)
        self.update(data)
        self.model.update(data)
        return self

    @classmethod
    def from_user(cls, user):
        """Create the user aggregate from the given user."""
        # TODO
        data = parse_user_data(user)

        model = cls.model_cls(data, model_obj=user)
        user_agg = cls(data, model=model)
        return user_agg

    @classmethod
    def get_record(cls, id_):
        """Get the user via the specified ID."""
        # TODO the the datastore.get_user() method will resolve both
        #      ID as well as email, which we do not necessarily want
        user = current_datastore.get_user(id_)
        if user is None:
            return None

        return cls.from_user(user)


class GroupAggregate(Record):
    """An aggregate of information about a user group/role."""

    model_cls = GroupAggregateModel
    """The model class for the user group aggregate."""

    # NOTE: the "uuid" isn't a UUID but contains the same value as the "id"
    #       field, which is currently a str for Role objects (role.name)!
    dumper = SearchDumper(extensions=[], model_fields={"id": ("uuid", str)})

    metadata = None
    """Disabled metadata field from the base class."""

    index = IndexField("groups-group-v1.0.0", search_alias="groups")
    """The search engine index to use."""

    # TODO
    id = ModelField("id")
    """The data-layer id."""

    name = DictField("name")
    """The group's name."""

    is_managed = DictField("is_managed")
    """If the group is managed manually."""

    _role = None
    """The cached Role entity."""

    @property
    def avatar_chars(self):
        """Get avatar characters for user."""
        return self.id[0].upper()

    @property
    def avatar_color(self):
        """Get avatar color for user."""
        colors = current_app.config["USERS_RESOURCES_AVATAR_COLORS"]
        normalized_group_initial = unicodedata.normalize("NFKD", self.id[0]).encode(
            "ascii", "ignore"
        )
        return colors[int(normalized_group_initial, base=36) % len(colors)]

    @property
    def role(self):
        """Cache for the associated role object."""
        role = self._role
        if role is None:
            if role is None and self.id is not None:
                role = current_datastore.find_role(self.id)

            self._role = role

        return role

    def commit(self):
        """Update the aggregate data on commit."""
        # TODO this does not allow us to set properties via the aggregate?
        #      because everything's taken from the Role object...
        data = parse_role_data(self.role)
        self.update(data)
        self.model.update(data)
        return self

    @classmethod
    def from_role(cls, role):
        """Create the user group aggregate from the given role."""
        # TODO
        data = parse_role_data(role)

        model = cls.model_cls(data, model_obj=role)
        role_agg = cls(data, model=model)
        role_agg._role = role
        return role_agg

    @classmethod
    def get_record(cls, id_):
        """Get the user group via the specified ID."""
        # TODO how do we want to resolve the roles? via ID or name?
        role = current_datastore.role_model.query.get(id_)
        if role is None:
            return None

        return cls.from_role(role)

    @classmethod
    def get_record_by_name(cls, name):
        """Get the user group via the specified ID."""
        # TODO how do we want to resolve the roles? via ID or name?
        role = current_datastore.role_model.query.filter_by(name=name).one_or_none()
        if role is None:
            return None

        return cls.from_role(role)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User service tests."""

import pytest
from invenio_records_resources.services.errors import PermissionDeniedError


#
# Search
#
def test_search_restricted(user_service, anon_identity, user_pub):
    """Only authenticated users can search."""
    # Anon identity
    pytest.raises(
        # TODO: Should be mapped to a 404
        PermissionDeniedError,
        user_service.search,
        anon_identity,
    )
    # Authenticated identity
    res = user_service.search(user_pub.identity).to_dict()
    assert res["hits"]["total"] > 0


def test_search_public_users(user_service, user_pub):
    """Only public users are shown in search."""
    res = user_service.search(user_pub.identity).to_dict()
    assert res["hits"]["total"] == 2  # 2 public users in conftest


@pytest.mark.parametrize(
    "query",
    [
        "email:res@inveniosoftware.org",
        "res@inveniosoftware.org",
        "email:pubres@inveniosoftware.org",
        "pubres@inveniosoftware.org",
        "Plazi",
        "+name:Jose -affiliation:CERN",
        "name:Jose AND NOT affiliation:CERN",
        "username:inactive",
        "username:unconfirmed",
        "preferences.visibility:public",
        "preferences.email_visibility:restricted",
        "profile.affiliations:Plazi",
        "invalid:test",
    ],
)
def test_search_field_not_searchable(user_service, user_pub, query):
    """Make sure certain fields are NOT searchable."""
    res = user_service.search(user_pub.identity, q=query).to_dict()
    assert res["hits"]["total"] == 0


@pytest.mark.parametrize(
    "query",
    [
        "affiliations:CERN",
        "affiliation:CERN",
        "name:Jose affiliation:CERN",
        "+name:Jose +affiliation:CERN",
        "CERN",
        "Tim",
        "Tim CERN",
        "Jose",
        "Jose CERN",
        "email:pub@inveniosoftware.org",
        "username:pub",
    ],
)
def test_search_field(user_service, user_pub, query):
    """Make sure certain fields ARE searchable."""
    res = user_service.search(user_pub.identity, q=query).to_dict()
    assert res["hits"]["total"] > 0


#
# Read
#
def test_read_with_anon(user_service, anon_identity, user_pub, user_pubres, user_res):
    """Anonymous users can read a single *public* user."""
    res = user_service.read(anon_identity, user_pub.id).to_dict()
    assert res["username"] == "pub"
    assert res["email"] == user_pub.email
    res = user_service.read(anon_identity, user_pubres.id).to_dict()
    assert res["username"] == "pubres"
    assert "email" not in res
    pytest.raises(
        # TODO: Should be mapped to a 404
        PermissionDeniedError,
        user_service.read,
        anon_identity,
        user_res.id,
    )


@pytest.mark.parametrize(
    "username,can_read", [("pub", True), ("pubres", True), ("res", False)]
)
def test_read_avatar_with_anon(user_service, anon_identity, users, username, can_read):
    """Anonymous users can read avatar single *public* user."""
    user = users[username]
    if can_read:
        user_service.read_avatar(anon_identity, user.id)
    else:
        pytest.raises(
            # TODO: Should be mapped to a 404
            PermissionDeniedError,
            user_service.read_avatar,
            anon_identity,
            user.id,
        )


@pytest.mark.parametrize(
    "username",
    [
        "pub",
        "pubres",
        "res",
    ],
)
def test_read_self(user_service, users, username):
    """Users can read themselves."""
    user = users[username]
    res = user_service.read(user.identity, user.id).to_dict()
    assert res["username"] == username
    # Email can be seen even if restricted
    assert res["email"] == user.email
    # Avatar can also be seen.
    user_service.read_avatar(user.identity, user.id)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users service configuration."""

from invenio_records_resources.services import (
    RecordServiceConfig,
    SearchOptions,
    pagination_links,
)
from invenio_records_resources.services.records.params import QueryStrParam, SortParam
from invenio_records_resources.services.records.queryparser import (
    QueryParser,
    SearchFieldTransformer,
)

from ...records.api import UserAggregate
from ..common import Link
from ..params import FixedPagination
from ..permissions import UsersPermissionPolicy
from ..schemas import UserSchema
from .results import UserItem, UserList


class UserSearchOptions(SearchOptions):
    """Search options."""

    pagination_options = {
        "default_results_per_page": 10,
        "default_max_results": 10,
    }

    query_parser_cls = QueryParser.factory(
        fields=["username^2", "email^2", "profile.full_name^3", "profile.affiliations"],
        tree_transformer_factory=SearchFieldTransformer.factory(
            mapping={
                "affiliation": "profile.affiliations",
                "affiliations": "profile.affiliations",
                "email": "email",
                "full_name": "profile.full_name",
                "fullname": "profile.full_name",
                "name": "profile.full_name",
                "username": "username",
            }
        ),
    )

    params_interpreters_cls = [
        QueryStrParam,
        SortParam,
        FixedPagination,
    ]


class UsersServiceConfig(RecordServiceConfig):
    """Requests service configuration."""

    # common configuration
    permission_policy_cls = UsersPermissionPolicy
    result_item_cls = UserItem
    result_list_cls = UserList
    search = UserSearchOptions

    # specific configuration
    service_id = "users"
    record_cls = UserAggregate
    schema = UserSchema
    indexer_queue_name = "users"
    index_dumper = None

    # links configuration
    links_item = {
        "self": Link("{+api}/users/{id}"),
        "avatar": Link("{+api}/users/{id}/avatar.svg"),
    }
    links_search = pagination_links("{+api}/users{?args*}")

    components = [
        # order of components are important!
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users and user groups permissions."""

from invenio_records_permissions import BasePermissionPolicy
from invenio_records_permissions.generators import (
    AnyUser,
    AuthenticatedUser,
    SystemProcess,
)

from .generators import IfPublicEmail, IfPublicUser, Self


class UsersPermissionPolicy(BasePermissionPolicy):
    """Permission policy for users and user groups."""

    can_create = [SystemProcess()]
    can_read = [IfPublicUser([AnyUser()], [Self()]), SystemProcess()]
    can_search = [AuthenticatedUser(), SystemProcess()]
    can_update = [SystemProcess()]
    can_delete = [SystemProcess()]

    can_read_email = [IfPublicEmail([AnyUser()], [Self()]), SystemProcess()]
    can_read_details = [Self(), SystemProcess()]


class GroupsPermissionPolicy(BasePermissionPolicy):
    """Permission policy for users and user groups."""

    can_create = [SystemProcess()]
    can_read = [AuthenticatedUser(), SystemProcess()]
    can_search = [AuthenticatedUser(), SystemProcess()]
    can_update = [SystemProcess()]
    can_delete = [SystemProcess()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN
# Copyright (C) 2022 TU Wien.
#
# Invenio-Records-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver and proxy for users."""

from types import SimpleNamespace

from flask_principal import UserNeed
from invenio_accounts.models import User
from invenio_records_resources.references.resolvers import EntityProxy, EntityResolver

from .proxies import current_users_service
from .services.users.config import UsersServiceConfig


class UserProxy(EntityProxy):
    """Resolver proxy for a User entity."""

    def _resolve(self):
        """Resolve the User from the proxy's reference dict."""
        user_id = int(self._parse_ref_dict_id())
        return User.query.get(user_id)

    def get_needs(self, ctx=None):
        """Get the UserNeed for the referenced user."""
        user_id = int(self._parse_ref_dict_id())
        return [UserNeed(user_id)]

    def pick_resolved_fields(self, resolved_dict):
        """Select which fields to return when resolving the reference."""
        profile = resolved_dict.get("profile", {})
        fake_user_obj = SimpleNamespace(id=resolved_dict["id"])
        avatar = current_users_service.links_item_tpl.expand(fake_user_obj)["avatar"]
        return {
            "id": resolved_dict["id"],
            "username": resolved_dict["username"],
            "email": resolved_dict.get("email", ""),
            "profile": {
                "full_name": profile.get("full_name", ""),
                "affiliations": profile.get("affiliations", ""),
            },
            "links": {"avatar": avatar},
        }


class UserResolver(EntityResolver):
    """Resolver for users."""

    type_id = "user"

    def __init__(self):
        """Constructor."""
        super().__init__(UsersServiceConfig.service_id)

    def matches_reference_dict(self, ref_dict):
        """Check if the reference dict references a user."""
        return self._parse_ref_dict_type(ref_dict) == self.type_id

    def _reference_entity(self, entity):
        """Create a reference dict for the given user."""
        return {"user": str(entity.id)}

    def matches_entity(self, entity):
        """Check if the entity is a user."""
        return isinstance(entity, User)

    def _get_entity_proxy(self, ref_dict):
        """Return a UserProxy for the given reference dict."""
        return UserProxy(self, ref_dict)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module providing management APIs for users and roles/groups."""


USERS_RESOURCES_AVATAR_COLORS = [
    "#e06055",
    "#ff8a65",
    "#e91e63",
    "#f06292",
    "#673ab7",
    "#ba68c8",
    "#7986cb",
    "#3f51b5",
    "#5e97f6",
    "#00a4e4",
    "#4dd0e1",
    "#0097a7",
    "#d4e157",
    "#aed581",
    "#57bb8a",
    "#4db6ac",
    "#607d8b",
    "#795548",
    "#a1887f",
    "#fdd835",
    "#a3a3a3",
    "#556c60",
    "#605264",
    "#923035",
    "#915a30",
    "#55526f",
    "#67635a",
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users service tasks."""

from celery import shared_task
from flask import current_app
from invenio_records_resources.tasks import send_change_notifications
from invenio_search.engine import search

from ...proxies import current_users_service
from ...records.api import UserAggregate


@shared_task(ignore_result=True)
def reindex_user(user_id):
    """Reindex the given user."""
    index = current_users_service.record_cls.index
    if current_users_service.indexer.exists(index):
        try:
            user_agg = UserAggregate.get_record(user_id)
            current_users_service.indexer.index(user_agg)
            # trigger reindexing of related records
            send_change_notifications(
                "users", [(user_agg.id, str(user_agg.id), user_agg.revision_id)]
            )
        except search.exceptions.ConflictError as e:
            current_app.logger.warn(f"Could not reindex user {user_id}: {e}")


@shared_task(ignore_result=True)
def unindex_user(user_id):
    """Delete the given user from the index."""
    index = current_users_service.record_cls.index
    if current_users_service.indexer.exists(index):
        try:
            user_agg = UserAggregate.get_record(user_id)
            current_users_service.indexer.delete(user_agg)
        except search.exceptions.ConflictError as e:
            current_app.logger.warn(f"Could not unindex user {user_id}: {e}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 KTH Royal Institute of Technology
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 European Union.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users service."""

from invenio_accounts.models import User
from invenio_records_resources.resources.errors import PermissionDeniedError
from invenio_records_resources.services import RecordService
from invenio_records_resources.services.uow import RecordCommitOp, unit_of_work
from invenio_search.engine import dsl

from invenio_users_resources.services.results import AvatarResult

from ...records.api import UserAggregate


class UsersService(RecordService):
    """Users service."""

    @property
    def user_cls(self):
        """Alias for record_cls."""
        return self.record_cls

    @unit_of_work()
    def create(self, identity, data, raise_errors=True, uow=None):
        """Create a user."""
        self.require_permission(identity, "create")

        # validate data
        data, errors = self.schema.load(
            data,
            context={"identity": identity},
        )

        # create the user with the specified data
        user = self.user_cls.create(data)

        # run components
        self.run_components(
            "create",
            identity,
            data=data,
            user=user,
            errors=errors,
            uow=uow,
        )

        # persist user to DB (indexing is done in the session hooks, see ext)
        uow.register(RecordCommitOp(user))

        return self.result_item(
            self, identity, user, links_tpl=self.links_item_tpl, errors=errors
        )

    def search(self, identity, params=None, search_preference=None, **kwargs):
        """Search for records matching the querystring."""
        return super().search(
            identity,
            params=params,
            search_preference=search_preference,
            extra_filter=dsl.Q("term", active=True) & dsl.Q("term", confirmed=True),
            **kwargs,
        )

    def read(self, identity, id_):
        """Retrieve a user."""
        # resolve and require permission
        user = UserAggregate.get_record(id_)
        if user is None:
            # return 403 even on empty resource due to security implications
            raise PermissionDeniedError()

        self.require_permission(identity, "read", record=user)

        # run components
        for component in self.components:
            if hasattr(component, "read"):
                component.read(identity, user=user)

        return self.result_item(self, identity, user, links_tpl=self.links_item_tpl)

    def read_avatar(self, identity, id_):
        """Get a user's avatar."""
        user = UserAggregate.get_record(id_)
        if user is None:
            # return 403 even on empty resource due to security implications
            raise PermissionDeniedError()
        self.require_permission(identity, "read", record=user)
        return AvatarResult(user)

    def rebuild_index(self, identity, uow=None):
        """Reindex all users managed by this service."""
        for user in User.query.all():
            user_agg = self.record_cls.from_user(user)
            self.indexer.index(user_agg)

        return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio users DB hooks."""

from invenio_accounts.models import Role, User
from invenio_accounts.proxies import current_db_change_history

from ..services.groups.tasks import reindex_group, unindex_group
from ..services.users.tasks import reindex_user, unindex_user


def pre_commit(sender, session):
    """Find out which entities need indexing before commit."""
    # it seems that the {dirty,new,deleted} sets aren't populated
    # in after_commit anymore, that's why we need to collect the
    # information here.
    # session is a scoped_session and does not have _model_changes
    # so we have rely only on .dirty
    updated = session.dirty.union(session.new)
    deleted = session.deleted
    sid = id(session)
    # flush the session s.t. related models are queryable
    session.flush()

    # users need to be reindexed if their user model was updated, or
    # their profile was changed (or even possibly deleted)
    current_db_change_history.updated_users[sid].extend(
        [u.id for u in updated if isinstance(u, User)]
    )
    current_db_change_history.updated_roles[sid].extend(
        [r.id for r in updated if isinstance(r, Role)]
    )

    current_db_change_history.deleted_users[sid].extend(
        [u.id for u in deleted if isinstance(u, User)]
    )
    current_db_change_history.deleted_roles[sid].extend(
        [r.id for r in deleted if isinstance(r, Role)]
    )


def post_commit(sender, session):
    """Reindex all modified users and roles after the DB commit."""
    # since this function is called after the commit, no more
    # DB operations are allowed here, not even lazy-loading of
    # properties!
    sid = id(session)
    for user_id in current_db_change_history.updated_users[sid]:
        reindex_user.delay(user_id)

    for role_id in current_db_change_history.updated_roles[sid]:
        reindex_group.delay(role_id)

    for user_id in current_db_change_history.deleted_users[sid]:
        unindex_user.delay(user_id)
    for role_id in current_db_change_history.deleted_roles[sid]:
        unindex_group.delay(role_id)

    current_db_change_history._clear_dirty_sets(session)
